--------------------------------------------------------------------------------
-- Copyright (C) 2004 Trenz Electronic GmbH
-- www.trenz-electronic.de
-- Author:      Kolja Sulimma
--------------------------------------------------------------------------------
-- Project:     TE-XC3S Application Note: Pong Demo
-- File:        toplevel.vhd
-- Description: Pong Demo
-- History:     V1.0 2004-10-11 KS created
--------------------------------------------------------------------------------
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.

-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity toplevel is
    Port ( 
	 	-- the following output assignments are required so that
		-- the GT3200 USB phy generates a 30 MHz clock
		utmi_databus16_8  : out bit;
		utmi_reset			: out bit;
		utmi_xcvrselect	: out bit;
		utmi_termselect	: out bit;
		utmi_opmode1		: out bit;
		utmi_txvalid		: out bit;

		-- this is the 30 MHz clock input (clkout is the utmi name)
		utmi_clkout			: in std_logic;

		--vga
		fpga_r : out std_logic_vector(2 downto 0);
		fpga_g : out std_logic_vector(2 downto 0);
		fpga_b : out std_logic_vector(2 downto 0);
		vsync_b, hsync_b : inout std_logic;

		--keyboard interfaces
		ps2_clk1, ps2_data1 : in std_logic;
		ps2_clk2, ps2_data2 : in std_logic;

		--leds on baseboard
		led : out std_logic_vector(1 to 4);

		--joystick and buttons
		joy_up, joy_down, joy_right, joy_left, joy_fire : in std_logic; 

		 --reset is connected to R16 and pulled up
		reset : in std_logic;

		--led	on micromodule
	 	mm_led : out std_logic;

		--audio
		aud_out : out std_logic_vector(1 to 4);

		--sram
		ram_a : out std_logic_vector(18 downto 1);
		ram_io : inout std_logic_vector(15 downto 0);
		ram_ce : out std_logic;
		ram_oe : out std_logic;
		ram_we : out std_logic;
		ram_bhe : out std_logic;
		ram_ble : out std_logic;

		--rs232 
		fpga_txd, fpga_rts : out std_logic;
		fpga_rxd, fpga_cts : in std_logic;

		--ir remote control
		ir_data : in std_logic;

		--lcd_display
		lcd_rs: out std_logic;
		lcd_rw: out std_logic;
		lcd_e: out std_logic;
		lcd_d: inout std_logic_vector(3 downto 0);

		--compact flash
		cf_we: out std_logic;		  -- all these signals are active low
		cf_reg: out std_logic;		  -- for more details see the specification
		cf_cs0: out std_logic;		  -- of compact flash
		cf_cs1: out std_logic;
		cf_reset: out std_logic;
		cf_iord: out std_logic;
		cf_iowr: out std_logic;
		cf_irq: in std_logic;
		cf_wait: in std_logic;
		cf_cd1: in std_logic;
		cf_cd2: in std_logic;
		iois16: in std_logic;
		cf_oe: out std_logic;
		cf_dasp: inout std_logic;
		cf_pdiag: inout std_logic;

		--cf power enable (active low)
		cf_pwr_en: out std_logic

		
	 );
end toplevel;

architecture Behavioral of toplevel is
   -- generate vga timing for a resolution of 512x384 pixels
	-- the origin is on the lower right edge of the screen
	component timing1024x384 is 
		port (clk : in std_logic; 
				x :out std_logic_vector(9 downto 0);
				y : out std_logic_vector(8 downto 0);
				vsync_b, hsync_b, blank : inout std_logic;
				framestrobe : out std_logic);
	end component;

   -- sprite components
	-- currently they are used to display colored rectangles
	-- using the hit signal, but the address signal can also
	-- be used to display a sprite image from a rom
   component sprite is
		Generic (
			  shift_x : integer := 0;	 -- scale in x coordinate by 2**shift_x
			  shift_y : integer := 0;	 -- scale in y coordinate by 2**shift_y
			  bits_x  : integer;		    -- screen size in x coordinate is 2**(shift_x + bits_x)
			  size_y  : integer			 -- screen size in y coordinate
	   );
   	Port ( 
			  screen_x : in std_logic_vector(8 downto 0);   -- position cathode ray on screen
           screen_y : in std_logic_vector(8 downto 0);
           pos_x : in std_logic_vector(8 downto 0);	   -- lower right corner of sprite on screen
           pos_y : in std_logic_vector(8 downto 0);
           hit : out std_logic;								   -- '1' if cathode ray is inside sprite rectangle 
           address : out std_logic_vector(14 downto 0)); -- relative address of current sprite pixel
			  --address is the relative address in memory
			  --to the beginning of the sprite data
	end component;

	component ps2 is
	  -- this component processes the PS2 keyboard protocol and keep track whether certain keys are pressed.
	  -- currently the keys are cursor righ, left, up and down, keypad 0 and space but this can easily be changed.
	  -- this component can handle any number of simultaneausly pressed keys, but the keyboards usually only support two.
     Port (clk : in std_logic;
	 		  ps2_clk : in std_logic;
			  ps2_data : in std_logic;
			  right, left, up, down, kp0, space : out std_logic
			 );
   end component;

   component sram is 
 	  -- sram
	  -- The sram is not used by this demo. Instead a LFSR test is implemented.
	  -- Pseudo random numbers are written to the ram and read back. If a mismatch
	  -- is detected one of the LEDs is turned off.
   port(
		clk : in std_logic;
		reset : in std_logic;
		led : out std_logic_vector(1 to 4);

		ram_a : out std_logic_vector(18 downto 1);
		ram_io : inout std_logic_vector(15 downto 0);
		ram_ce : out std_logic;
		ram_oe : out std_logic;
		ram_we : out std_logic;
		ram_bhe : out std_logic;
		ram_ble : out std_logic );
   end component;

   component sound is
   	  -- sound output 
	  -- this component outputs a beep every time the ball is hit
	  -- y(3) toggles with about 6kHz, so we used it as a sound source
	  --	of course a real sigma delta converter with a small synthesizer would
	  -- sound a lot better. See http://www.xilinx.com/bvdocs/appnotes/xapp154.pdf
    Port ( clk : in std_logic;
    		sound_clk : in std_logic;
    		 start_beep : in std_logic;
           framestrobe : in std_logic;
           dir_x : in std_logic;
           aud_out : out std_logic_vector(1 to 4)	 );
	end component;

   component lcd is
   -- this component initializes the display and executes simple write commands
	-- for display output
	-- see lcd_use for an example
    Port ( clk,res : in std_logic;
    			  clk_25us,clk_50us: in std_logic;
	 		  write : in std_logic;
			  data : in std_logic_vector(7 downto 0);
			  ready : out std_logic;

			  lcd_data: inout std_logic_vector(3 downto 0);
			  lcd_rw: out std_logic;
			  lcd_rs: out std_logic;
			  lcd_en: out std_logic);
	end component;


	--This component was generated from clock_synthesis.xaw
	--It synthesizes a 40MHz clock from a 30MHz Clock

	--For bitstream generation the lck_clk option must be set to a value larger than 0
	--in the ISE GUI it is in the properties of "Generate Programming File"
	--in the startup options panel
	component clock_synthesis
	port(
		CLKIN_IN : IN std_logic;          
		CLKFX_OUT : OUT std_logic;
		CLKIN_IBUFG_OUT : OUT std_logic;
		CLK0_OUT : OUT std_logic
		);
	end component;

  
   --internal video timing signals signals
  	signal x : std_logic_vector(9 downto 0);
	signal y : std_logic_vector(8 downto 0);	-- position of the cathode ray
	signal blank, clk  : std_logic;	-- blank if outside visible screen area
	signal framestrobe : std_logic;  -- one clock cycle pulse for each frame
	signal rgb : std_logic_vector(8 downto 0); -- the vga color output

	--player controls 
	signal lleft, lright, lup, ldown, lfire, lspace : std_logic;  --left bat (joystick or 1st keyboard)
	signal rleft, rright, rup, rdown, rfire, rspace : std_logic;  --right bat (2nd keyboard)
  	signal kleft, kright, kup, kdown, kfire : std_logic;	--first keyboard, used to compute lleft, etc.

	--game control signals
	signal lbat_y, rbat_y : std_logic_vector(8 downto 0);	 -- position of the two bats
	signal lbat_x, rbat_x : std_logic_vector(9 downto 0);

	signal ball_x : std_logic_vector(9 downto 0);			 -- position of the ball
	signal ball_y : std_logic_vector(8 downto 0);	 
	signal dir_x, dir_y : std_logic;								 -- direction of the ball, (1, 1) is toward top left
	signal hit: std_logic_vector(0 to 2);						 -- cathode ray inside bat1, bat2, ball

   --audio control
	signal start_beep : std_logic;								 -- obvious ;-)
   
   --display control
	signal clk_25us,clk_50us: std_logic;
	signal ascii_data: std_logic_vector(7 downto 0);
	signal ascii_write: std_logic;
	signal lcd_ready: std_logic;

	type GAME_TYPE is (SPLASH, LSTART, RSTART, RUN);	    -- the state machine that controls the game
	signal gamestate : GAME_TYPE;
	constant ballspeedx : integer := 6;							 -- modifiy this depending on your skill
	constant ballspeedy : integer := 3;							 -- modifiy this depending on your skill
	constant batspeedx : integer := 8;
	constant batspeedy : integer := 4;							 
							 

begin

--configure utmi for 30MHz clock
utmi_databus16_8 	<= '1';
utmi_reset			<= '0';
utmi_xcvrselect	<= '1';
utmi_termselect	<= '1';
utmi_opmode1		<= '0';
utmi_txvalid		<= '0'; 

--configure compact flash for TRUE IDE mode
cf_pwr_en<= (cf_cd1) or  (cf_cd2);	  -- power enable when card detect
cf_oe<='0';					  -- TRUE IDE mode

cf_reg<='1';					  
cf_we<='1';
cf_iord<='1';
cf_iowr<='1';
cf_reset<='1';
cf_cs0<='1';
cf_cs1<='1';

--connect rs232 loopback
fpga_txd <= fpga_rxd;
fpga_rts <= fpga_cts;

-- debug output
--	input detected from an IR remote control is forwarded to the LED on the micromodule.
-- there it is easily accessible by a scope and you can also see some flicker when pressing a button.
mm_led <= ir_data;

--alias long names
fpga_r <= rgb(8 downto 6);
fpga_g <= rgb(5 downto 3);
fpga_b <= rgb(2 downto 0);

inst_clock_synthesis: clock_synthesis PORT MAP(
		CLKIN_IN => utmi_clkout,
		CLKFX_OUT => clk,
		CLKIN_IBUFG_OUT => open,
		CLK0_OUT => open
	);

-- component to create VGA timing
timing_module : timing1024x384			
	port map (clk=>clk, 
				x=>x, y=>y,
				vsync_b=>vsync_b, hsync_b=>hsync_b,
				blank=>blank, framestrobe=>framestrobe
				);

--sprites
left_bat : sprite
	generic map	(bits_x => 3, size_y => 64) 
	port map (
		screen_x => x(9 downto 1),
		screen_y => y,
		pos_x => lbat_x(9 downto 1),
		pos_y => lbat_y,
		hit => hit(0),
		address => open);
right_bat : sprite
	generic map	(bits_x => 3, size_y => 64) 
	port map (
		screen_x => x(9 downto 1),
		screen_y => y,
		pos_x => rbat_x(9 downto 1),
		pos_y => rbat_y,
		hit => hit(1),
		address => open);

ball : sprite
	generic map	(bits_x => 3, size_y => 8) 
	port map (
		screen_x => x(9 downto 1),
		screen_y => y,
		pos_x => ball_x(9 downto 1),
		pos_y => ball_y,
		hit => hit(2),
		address => open);

keyboard1 : ps2 
	port map (
		clk 		=> clk, 
		ps2_clk 	=> ps2_clk1,
		ps2_data => ps2_data1,
		right		=> kright,
		left		=> kleft,
		up			=> kup,
		down		=> kdown,
		kp0		=> kfire,
		space		=> lspace
		);

keyboard2 : ps2 
	port map (
		clk 		=> clk, 
		ps2_clk 	=> ps2_clk2,
		ps2_data => ps2_data2,
		right		=> rright,
		left		=> rleft,
		up			=> rup,
		down		=> rdown,
		kp0		=> rfire,
		space		=> rspace
		);

 -- memory test
sram_test: sram 
	port map(
		clk	=>	clk,
		reset => 	reset,
		led   =>	led,
		ram_a =>	ram_a,
		ram_io =>	ram_io,
		ram_ce => ram_ce,
		ram_oe =>	ram_oe,
		ram_we => ram_we,
		ram_bhe =>ram_bhe,
		ram_ble =>ram_ble 
		);
--beeps

sound_output: sound
	port map(
		clk	=>	clk,
		sound_clk => y(3),
		start_beep => start_beep,
		framestrobe => framestrobe,
		dir_x => dir_x,
		aud_out => aud_out
		);
		
 -- lcd display
lcd_set: lcd 
	port map (
			clk=>clk,
			res=>reset,
			clk_25us	=> clk_25us,
			clk_50us	=> clk_50us,
			write	=> ascii_write,
			data		=> ascii_data,
			lcd_data	=> lcd_d,
			lcd_rw	=> lcd_rw,
			lcd_en	=> lcd_e,
			lcd_rs	=> lcd_rs,
			ready	=> lcd_ready);

-- synchronize some inputs and invert joystick signals
-- (the joystick produces a '1' when a button is pressed)
-- instead of using an external joystick connected to J4 you
-- can use the switches S1 to S5 or a keyboard connected to J11
process 
begin
	wait until rising_edge(clk);
	lleft  	<= kleft  or not joy_left;
	lright 	<= kright or not joy_right;
	lup		<= kup	 or not joy_up;
	ldown		<= kdown	 or not joy_down;
	lfire		<= kfire  or not joy_fire; 
end process;



--game
process (clk,reset)
variable bounce : std_logic;
begin
  bounce := '0';
  if reset = '0' then
	 gamestate <= SPLASH;
  elsif rising_edge(clk) then
    case gamestate is 

	 	-- this state is named SPLASH because I intend to display a welcome image at this point
	 	when SPLASH =>
			gamestate <= LSTART;

		-- the right player has a ball and can start the game by pressing keypad 0
		when RSTART =>
			 dir_x <= '1';
			 if rfire = '1' then -- of course this should be the space key....
			 	gamestate <= RUN;
				bounce := '1';
 			 end if;

		-- the left player has a ball and can start the game by pressing his button
		when LSTART =>
			 dir_x <= '0';
			 if lfire = '1' then
			 	gamestate <= RUN;
				bounce := '1';
 			 end if;

		-- the game is running
		when RUN =>
		    --bouncing ball, use rectangular sprite collision detection
			 if hit(0) = '1' and hit(2) = '1' then	 --hit left bat
				dir_x <= '0';
				bounce := '1';
          elsif hit(1) = '1' and hit(2) = '1' then	--hit right bat
				dir_x <= '1';
				bounce := '1';
			 end if;

			 if ball_y < ballspeedy then --hit bottom
			   dir_y <= '1';
				bounce := '1';
			 elsif ball_y > 384 - ballspeedy then --hit top
			 	dir_y <= '0';
				bounce := '1';
          end if;
			   
			 if ball_x < ballspeedx then  -- ball left game to the right, left player could score a point
			 	gamestate <= RSTART;
			 elsif ball_x > (1023 - ballspeedx) then	-- ball left game to the left, right could score a point
			 	gamestate <= LSTART;
			 end if; 
		when others =>  --something went wrong, start over 
			gamestate <= SPLASH;
    end case;
  start_beep <= bounce;
  end if;
end process;

--ball coordinates
process(clk)
begin
  if rising_edge(clk) then
    case gamestate is
	   when SPLASH =>
		 	 ball_x <= "0000010000";
	 		 ball_y <= "011000000";
		when LSTART => --attach the ball to the left bat
			 ball_x <= lbat_x - 16;
			 ball_y <= lbat_y + 56;
		when RSTART =>	--attach the ball to the right bat
			 ball_x <= rbat_x + 16;
			 ball_y <= rbat_y + 56;
	   when RUN =>
		  if framestrobe = '1' then -- once per frame the ball coordinates are updated
		 	 if dir_x = '1' then
		   	ball_x <= ball_x + ballspeedx;
		 	 else
		  		ball_x <= ball_x - ballspeedx;
		 	 end if;
		 	 if dir_y = '1' then
		   	ball_y <= ball_y + ballspeedy;
		 	 else
		   	ball_y <= ball_y - ballspeedy;
		 	 end if;
	 	  end if;
	   when others =>
     end case;
  end if;
end process; 


--joystick control
process (clk,reset)
begin
  if reset = '0' then
	 rbat_y <= "010100000";
	 rbat_x <= "0000000000";
	 lbat_y <= "010100000";
	 lbat_x <= "1111110000";
  elsif rising_edge(clk) then
		if framestrobe = '1' then
		  --left bat up and down
		  if lup = '1' and ldown = '0' and lbat_y <= (384 - 64 - batspeedy) then
		  	  lbat_y <= lbat_y + batspeedy;
		  elsif lup = '0' and ldown = '1' and lbat_y >= batspeedy then
		  	  lbat_y <= lbat_y - batspeedy;
		  end if;  		
		   --left bat left and right
		  if lleft = '1' and lright = '0' and lbat_x < (1024 -16 -batspeedx) then
		  	  lbat_x <= lbat_x + batspeedx;
		  elsif lleft = '0' and lright = '1' and lbat_x > (1024 - 384 - batspeedx) then
		  	  lbat_x <= lbat_x - batspeedx;
		  end if;  	
		  --right bat up and down
		  if rup = '1' and rdown = '0' and rbat_y <= (384 - 64 - batspeedy) then
		  	  rbat_y <= rbat_y + batspeedy;
		  elsif rup = '0' and rdown = '1' and rbat_y >= batspeedy then
		  	  rbat_y <= rbat_y - batspeedy;
		  end if;  		
		   --left bat left and right
		  if rleft = '1' and rright = '0' and rbat_x < (384 - 16  - batspeedx) then
		  	  rbat_x <= rbat_x + batspeedx;
		  elsif rleft = '0' and rright = '1' and rbat_x > ( batspeedx) then
		  	  rbat_x <= rbat_x - batspeedx;
		  end if;  	
		end if;
	end if;
end process;


-- colors
process (clk) -- vga colors
begin
	if rising_edge(clk) then
		if blank = '1' then 
			rgb <= "000000000";
		else
		  if hit(0) = '1' then --left bat
		    rgb <= "100" & "111" & "000";
        elsif hit(1) = '1' then --right bat
		    rgb <= "100" & "000" & "111";
		  elsif hit(2) = '1' then --ball
		    rgb <= "111" & "000" & "000";
		  else 
		  	 rgb <= "000000010";  -- dark blue background
        end if;
		end if;
   end if;
end process;

-- this process output the string "Ping Pong" on the display by using modul lcd

 use_lcd: process(clk)
 		type lcd_data_type is (d1,d2,d3,d4,d5,d6,d7,d8,d9,d10);
 		variable lcd_data_state: lcd_data_type;
		begin
		  if rising_edge(clk)  then
		     if (reset='0') then lcd_data_state:=d1;
               elsif (framestrobe='1') then    -- this clocktime must be greater than 2x50us(2xclk_50us)
			 if  lcd_ready='1' then		  -- checks readyflag of lcd
			   
		   	   case lcd_data_state is			  -- writes character,one for one 
			  	when d1 =>   ascii_data<=x"50";	   --P
					        lcd_data_state:=d2;
						   ascii_write<='1';		  -- write command for lcd

				when d2 =>   ascii_data<=x"69";	   --i
					  	   lcd_data_state:=d3;
						   ascii_write<='1';

				when d3 =>   ascii_data<=x"6E";	   --n
					  	   lcd_data_state:=d4;
						   ascii_write<='1';

				when d4 =>   ascii_data<=x"67";	   --g
					 	   lcd_data_state:=d5;
						   ascii_write<='1';

				when d5 =>   ascii_data<=x"A0";	   --space
					 	   lcd_data_state:=d6;
						   ascii_write<='1';

				when d6 =>   ascii_data<=x"50";	   --P
					 	   lcd_data_state:=d7;
						   ascii_write<='1';

				when d7 =>   ascii_data<=x"6F";	   --o
					  	   lcd_data_state:=d8;
						   ascii_write<='1';

				when d8 =>   ascii_data<=x"6E";	   --n
					  	   lcd_data_state:=d9;
						   ascii_write<='1';

				when d9 =>   ascii_data<=x"67";	   --g
					   	   lcd_data_state:=d10;		    
						   ascii_write<='1';

 				when d10 =>null;
			   end case;
		    else ascii_write<='0';		-- resets write command
		   end if;
		  end if;
		end if;
end process;

  -- clock divider for lcd display control
 clk_div: process(clk)
		variable count: integer range 0 to 2000;
	 	begin
		
		  if rising_edge(clk)  then
		   
			if count=1000 then count:=count+1;  
				clk_25us<='1';					-- clock time 25us	
			elsif count=2000 then count:=0;	
				clk_25us<='1';					
				clk_50us<='1';					-- clock time 50us
			else count:=count+1;
				clk_25us<='0';
				clk_50us<='0';
			end if;
		  end if;
	  	end process;


end Behavioral;
